/* 
 *
 * This file contains a starting point for creating a BLE Peripheral application. 
 * It contains a single Service with two characteristics
 *   - A LED State characteristic which can be read and controlled via BLE
 *   - A Button State characteristic which sends notifications on Changes and can be read via BLE
 *  The application also has timers and interrupts to blink LEDs and can be woken up after
 *  going to sleep via a button press
 *
 * When running on the Nordic PCA10028 Dev board it has the following features:
 *  
 *  1.  Starts a BLE Peripheral and begins advertising 
 *  2.  Advertising times out after 60 seconds and the device is put to sleep
 *  3.  During advertising, LED 4 Blinks ON and OFF every second driven by an AppTimer
 *  4.  Button 1 Turns LED 1 ON when Pressed and OFF when Released
 *  5.  Button 4 wakes up the Device from Sleep mode when Pressed
 *	6.  LED 2 is controlled ON and OFF via BLE Write Commands on TAG Service LED Characteristic
 *  7.  Button 2 sends a BLE notification when pressed and it's state can be read over BLE
 * 
 *
 *  
 *
 */

#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include "nordic_common.h"
#include "nrf.h"
#include "app_error.h"
#include "nrf_gpio.h"
#include "nrf51_bitfields.h"
#include "ble.h"
#include "ble_hci.h"
#include "ble_srv_common.h"
#include "ble_advdata.h"
#include "ble_conn_params.h"
#include "app_scheduler.h"
#include "softdevice_handler.h"
#include "app_timer_appsh.h"
#include "app_gpiote.h"
#include "ble_sfst_tag.h"
#include "config.h"
#include "SEGGER_RTT.h"
#include "nrf_delay.h"



//this macro let's us use printf() functionality with the SEGGER RTT
//it requires SEGGER_RTT.h, stdio.h and SEGGER_RTT.c
#define RTT_PRINTF(...)\
do{\
   char str[64];\
	sprintf(str, __VA_ARGS__);\
	SEGGER_RTT_WriteString(0, str);\
}while(0)



#define IS_SRVC_CHANGED_CHARACT_PRESENT 0        /**< Include or not the service_changed characteristic. if not enabled, the server's database cannot be changed for the lifetime of the device*/

#define LED_4															24		//use Pin 24 for PCA10028, Use Pin 10 for SFST100 Rev3 RB0
#define SF_TAG_LED												22		//LED 2 is used for the TAG LED that is controlled via BLE messages
#define LED_1															21    //Pin 21 on PCA10028
#define BUTTON_4 													20    //Pin 20 on PCA10028
#define BUTTON_1													17    //Pin 17 on PCA10028
#define BUTTON_2													18	  //Pin 18
#define BTN_1_BITMASK											0x00020000  //GPIO Pin 17
#define BTN_2_BITMASK											0x00040000  //GPIO Pin 18

#define BLE_MANUF_ID_ULB 								0x4154
#define DEVICE_NAME                     "SplitFast TAG"                             /**< Name of device. Will be included in the advertising data. */

#define APP_ADV_INTERVAL                64                                          /**< The advertising interval (in units of 0.625 ms. This value corresponds to 40 ms). */
#define APP_ADV_TIMEOUT_IN_SECONDS      60                                         /**< The advertising timeout (in units of seconds). */

// YOUR_JOB: Modify these according to requirements.
#define APP_TIMER_PRESCALER             0                                           /**< Value of the RTC1 PRESCALER register. */
#define APP_TIMER_MAX_TIMERS            4                                          /**< Maximum number of simultaneously created timers. */
#define APP_TIMER_OP_QUEUE_SIZE         6                                           /**< Size of timer operation queues. */

//<DGH>  define a timer for handing blinking the LED
#define LED_BLINK_OFF_TIME							APP_TIMER_TICKS(1000, APP_TIMER_PRESCALER)  //stay off for 1000ms
#define LED_BLINK_ON_TIME							  APP_TIMER_TICKS(1000, APP_TIMER_PRESCALER)   //blink on for 100ms
static app_timer_id_t 									m_led_blink_timer_id, m_timer2_id;
static uint8_t     											LED_state;
static uint32_t													count;
#define	PIN_HIGH_EVENT									0
#define PIN_LOW_EVENT										1

//</DGH>

#define MIN_CONN_INTERVAL               MSEC_TO_UNITS(400, UNIT_1_25_MS)            /**< Minimum acceptable connection interval (0.5 seconds). */
#define MAX_CONN_INTERVAL               MSEC_TO_UNITS(650, UNIT_1_25_MS)           /**< Maximum acceptable connection interval (1 second). */
#define SLAVE_LATENCY                   0                                           /**< Slave latency. */
#define CONN_SUP_TIMEOUT                MSEC_TO_UNITS(4000, UNIT_10_MS)             /**< Connection supervisory timeout (4 seconds). */
#define FIRST_CONN_PARAMS_UPDATE_DELAY  APP_TIMER_TICKS(5000, APP_TIMER_PRESCALER)  /**< Time from initiating event (connect or start of notification) to first time sd_ble_gap_conn_param_update is called (5 seconds). */
#define NEXT_CONN_PARAMS_UPDATE_DELAY   APP_TIMER_TICKS(30000, APP_TIMER_PRESCALER) /**< Time between each call to sd_ble_gap_conn_param_update after the first call (30 seconds). */
#define MAX_CONN_PARAMS_UPDATE_COUNT    3                                           /**< Number of attempts before giving up the connection parameter negotiation. */

#define APP_GPIOTE_MAX_USERS            2                                          /**< Maximum number of users of the GPIOTE handler. */
static app_gpiote_user_id_t							m_gpiote_user_1;
static void handle_gpiote_event(uint32_t event_pins_low_to_high, uint32_t event_pins_high_to_low); //prototype


//#define BUTTON_DETECTION_DELAY          APP_TIMER_TICKS(50, APP_TIMER_PRESCALER)    /**< Delay from a GPIOTE event until a button is reported as pushed (in number of timer ticks). */

#define SEC_PARAM_TIMEOUT	  						30																					// Timeout for Pairing request or Security request (in seconds)	
#define SEC_PARAM_BOND                  1                                           /**< Perform bonding. */
#define SEC_PARAM_MITM                  0                                           /**< Man In The Middle protection not required. */
#define SEC_PARAM_IO_CAPABILITIES       BLE_GAP_IO_CAPS_NONE                        /**< No I/O capabilities. */
#define SEC_PARAM_OOB                   0                                           /**< Out Of Band data not available. */
#define SEC_PARAM_MIN_KEY_SIZE          7                                           /**< Minimum encryption key size. */
#define SEC_PARAM_MAX_KEY_SIZE          16                                          /**< Maximum encryption key size. */

#define DEAD_BEEF                       0xDEADBEEF                                  /**< Value used as error code on stack dump, can be used to identify stack location on stack unwind. */

static ble_gap_sec_params_t             m_sec_params;                               /**< Security requirements for this application. */
static uint16_t                         m_conn_handle = BLE_CONN_HANDLE_INVALID;    /**< Handle of the current connection. */
static ble_tag_t												m_tag;																			//TAG Service data structure

// YOUR_JOB: Modify these according to requirements (e.g. if other event types are to pass through the scheduler).
#define SCHED_MAX_EVENT_DATA_SIZE       sizeof(app_timer_event_t)                   /**< Maximum size of scheduler events. Note that scheduler BLE stack events do not contain any data, as the events are being pulled from the stack in the event handler. */
#define SCHED_QUEUE_SIZE                10                                          /**< Maximum number of events in the scheduler queue. */

static uint16_t  debug_counter = 0;




//  This is the callback function that will be executed each time the LED blinking timer expires
//  The timer is a single shot, so we restart it here and give it a new timeout value. This allows
//  us to create timers that have different on and off durations
static void led_blink_timer_handler(void *p_context)
{
	
    uint32_t err_code, next_timer_interval; 

		if(LED_state)
		{
			 nrf_gpio_pin_set(LED_4);  //turn off the LED
			 nrf_gpio_pin_clear(10);   //RB0 low
			 next_timer_interval = LED_BLINK_OFF_TIME;   //stay off for OFF_TIME 
		}
		else
		{
			 nrf_gpio_pin_clear(LED_4);  //turn on the LED
			 nrf_gpio_pin_set(10);       //RB0 high
			 next_timer_interval = LED_BLINK_ON_TIME;   //turn on for ON_TIME 
		}
		LED_state = !LED_state;	
    
	  err_code = app_timer_start(m_led_blink_timer_id, next_timer_interval, NULL);  //reload the timer
    APP_ERROR_CHECK(err_code); 

}


static void timer2_handler(void *p_context)
{
   count++;
}


static void sendButtonChangeNotification(uint8_t state)
{
	uint32_t err_code;
	
	err_code = ble_tag_on_button_change(&m_tag, state);
	if(err_code != NRF_SUCCESS && err_code != BLE_ERROR_INVALID_CONN_HANDLE && err_code != NRF_ERROR_INVALID_STATE)
	{
		APP_ERROR_CHECK(err_code);
	}
	
}


void handle_gpiote_event(uint32_t event_pins_low_to_high, uint32_t event_pins_high_to_low)
{ 
	  if(event_pins_low_to_high & BTN_1_BITMASK)  //BTN 1 transitioned from Low to High = BTN released
		{
				nrf_gpio_pin_set(LED_1); //turn off LED_1
		}

	  if(event_pins_high_to_low & BTN_1_BITMASK)  //BTN 2 transitioned from High to Low = BTN pressed
		{
				nrf_gpio_pin_clear(LED_1);  //turn on LED_1
			  RTT_PRINTF("BTN-1 Count = %d\n", ++debug_counter);
		}	
		
		if(m_conn_handle != BLE_CONN_HANDLE_INVALID)
		{
		
				if(event_pins_low_to_high & BTN_2_BITMASK)  //BTN 1 transitioned from Low to High = BTN released
				{
					sendButtonChangeNotification(0);	
					nrf_gpio_pin_set(SF_TAG_LED); //turn off LED_2
					
				}
				
				if(event_pins_high_to_low & BTN_2_BITMASK)  //BTN 2 transitioned from High to Low = BTN pressed
				{
					sendButtonChangeNotification(1);	
					nrf_gpio_pin_clear(SF_TAG_LED);  //turn on LED_2
						SEGGER_RTT_WriteString(0, "BTN-2\n");  
				}	
		}
}


//Initialize the I/O pins
static void gpio_init(void)
{
	  
	  nrf_gpio_cfg_output(LED_4);
	  nrf_gpio_pin_clear(LED_4);     //Setting the pin will turn OFF the LED at startup
	
	  nrf_gpio_cfg_output(LED_1);
	  nrf_gpio_pin_set(LED_1);       //Setting the pin will turn OFF the LED at startup
	
	  nrf_gpio_cfg_output(SF_TAG_LED);
	  nrf_gpio_pin_set(SF_TAG_LED);  //Setting the pin will turn OFF the LED at startup
	
	
	
	  nrf_gpio_cfg_output(10);      //RB0 on SPLITFAST REV3 PCB
	
	  //On PCA10028 pressing a button connects the GPIO pin to GND so we need to configure a pullup resistor
    //to pull the pin high when the button is not pressed  
	  nrf_gpio_cfg_input(BUTTON_4, NRF_GPIO_PIN_PULLUP);
	  nrf_gpio_cfg_input(BUTTON_1, NRF_GPIO_PIN_PULLUP);
	  nrf_gpio_cfg_input(BUTTON_2, NRF_GPIO_PIN_PULLUP);
	
	  //this pin is used to wake up from system off. Set the Pin to sense when it goes low 
	  NRF_GPIO->PIN_CNF[BUTTON_4] |= (GPIO_PIN_CNF_SENSE_Low << GPIO_PIN_CNF_SENSE_Pos);
}



/* Callback function for asserts in the SoftDevice.
 *
 * @details This function will be called in case of an assert in the SoftDevice.
 *
 * @warning This handler is an example only and does not fit a final product. You need to analyze
 *          how your product is supposed to react in case of Assert.
 * @warning On assert from the SoftDevice, the system can only recover on reset.
 *
 * @param[in]   line_num   Line number of the failing ASSERT call.
 * @param[in]   file_name  File name of the failing ASSERT call.
 */
void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(DEAD_BEEF, line_num, p_file_name);
}



/*  Function for the Timer initialization.
 */
static void timers_init(void)
{
    uint32_t err_code;  
	  
	  // Initialize timer module, making it use the scheduler
    APP_TIMER_APPSH_INIT(APP_TIMER_PRESCALER, APP_TIMER_MAX_TIMERS, APP_TIMER_OP_QUEUE_SIZE, true);

    //Initialize the LED blinking timer
	  LED_state = 0;
    err_code = app_timer_create(&m_led_blink_timer_id, APP_TIMER_MODE_SINGLE_SHOT, led_blink_timer_handler);
    APP_ERROR_CHECK(err_code);
	
	  err_code = app_timer_create(&m_timer2_id, APP_TIMER_MODE_SINGLE_SHOT, timer2_handler);
    APP_ERROR_CHECK(err_code);
}


/* Function for the GAP initialization.
 *
 * This function sets up all the necessary GAP (Generic Access Profile) parameters of the
 * device including the device name, appearance, and the preferred connection parameters.
 */
static void gap_params_init(void)
{
    uint32_t                err_code;
    ble_gap_conn_params_t   gap_conn_params;
    ble_gap_conn_sec_mode_t sec_mode;

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&sec_mode);

    err_code = sd_ble_gap_device_name_set(&sec_mode, (const uint8_t *)DEVICE_NAME, strlen(DEVICE_NAME));
    APP_ERROR_CHECK(err_code);

    memset(&gap_conn_params, 0, sizeof(gap_conn_params));
    gap_conn_params.min_conn_interval = MIN_CONN_INTERVAL;
    gap_conn_params.max_conn_interval = MAX_CONN_INTERVAL;
    gap_conn_params.slave_latency     = SLAVE_LATENCY;
    gap_conn_params.conn_sup_timeout  = CONN_SUP_TIMEOUT;

    err_code = sd_ble_gap_ppcp_set(&gap_conn_params);
    APP_ERROR_CHECK(err_code);
}




//Simplified version of advertising_init() for debugging
static void advertising_init(void)
{
    uint32_t        err_code;
    ble_advdata_t   advdata;			//initial advertising data packet
	  ble_advdata_t   scanrsp;      //scan response advertising data packet
    

	/****PROBLEM IS HERE****/
	
    //ble_uuid_t adv_uuids[] = {{TAG_UUID_SERVICE, m_tag.uuid_type}};  //This should work but does not, causes the device to continuously reset
	  ble_uuid_t adv_uuids[] = {{TAG_UUID_SERVICE, BLE_UUID_TYPE_BLE}}; //this does work, but crashes when sending a notification
		
	
	  //build and set advdata
	  memset(&advdata, 0, sizeof(advdata));
		advdata.name_type               = BLE_ADVDATA_FULL_NAME;
    advdata.include_appearance      = true;
    advdata.flags                   = BLE_GAP_ADV_FLAGS_LE_ONLY_LIMITED_DISC_MODE;


		memset(&scanrsp, 0, sizeof(scanrsp));
 		scanrsp.uuids_complete.uuid_cnt = sizeof(adv_uuids) / sizeof(adv_uuids[0]);
		scanrsp.uuids_complete.p_uuids = adv_uuids;
		
		err_code = ble_advdata_set(&advdata, &scanrsp);
    APP_ERROR_CHECK(err_code);

	}



/*********************************************************************************

	WE"RE HANGING IN advertising_init() SO COMMENT THIS VERSION OF THE FUNCTION OUT FOR NOW
	SEE ABOVE FOR A SIMPLIFIED VERSION OF advertising_init() FOR DEBUGGING

 Function for initializing the Advertising functionality.
 *
 * Encodes the required advertising data and passes it to the stack.
 * Also builds a structure to be passed to the stack when starting advertising.
 
static void advertising_init(void)
{
    uint32_t        err_code;
    ble_advdata_t   advdata;						//initial advertising data packet
	  ble_advdata_t   advdata_response;   //scan response advertising data packet
    uint8_t         flags = BLE_GAP_ADV_FLAGS_LE_ONLY_LIMITED_DISC_MODE;
	

    //Advertise the ID for the TAG Service
		
	  //This should work but does not
		//ble_uuid_t adv_uuids[] = {{TAG_UUID_SERVICE, m_tag.uuid_type}};
	
	  //this does work
	  ble_uuid_t adv_uuids[] = {{TAG_UUID_SERVICE, BLE_UUID_TYPE_BLE}};
	
 		
		//vendor specific data
		ble_advdata_manuf_data_t     manuf_data;          //variable to hold custom data to be sent in intital advertising packet
				
		uint8_t data[6];
		data[0] = 0x01;		//TAG Model
		data[1] = 0x0A;		//TAG Version
		data[2] = 0x00;		//TAG Uniquie ID byte 4 MSB
		data[3] = 0x00;	  //TAG Uniquie ID byte 3
		data[4] = 0x00;   //TAG Uniquie ID byte 2
		data[5] = 0x01;   //TAG Uniquie ID byte 1 LSB
		
		manuf_data.company_identifier = BLE_MANUF_ID_ULB; 
		manuf_data.data.p_data 				= data;
		manuf_data.data.size					= 6;
		
		// Configure the advertising data
    memset(&advdata, 0, sizeof(advdata));

    advdata.name_type               = BLE_ADVDATA_SHORT_NAME;
		advdata.short_name_len					= 11;
    advdata.include_appearance      = false;
    advdata.flags                   = flags;
    advdata.uuids_complete.uuid_cnt = sizeof(adv_uuids) / sizeof(adv_uuids[0]);
    advdata.uuids_complete.p_uuids  = adv_uuids;
		advdata.p_manuf_specific_data   = &manuf_data;
		
		
		//Manufacturer specific scan response data
		ble_advdata_manuf_data_t       manuf_data_response; //variable to hold custom data to sent in response to more Adv data request
		
		uint8_t data_response[22];   //20 bytes of data available for response message Athlete ID=4bytes, Athlete Name=16bytes, Null term=2bytes
		data_response[0] = 0x00;			//Ath ID byte 4 MSB
		data_response[1] = 0x00;			//Ath ID byte 3 
		data_response[2] = 0x00;			//Ath ID byte 2 
		data_response[3] = 0x01;			//Ath ID byte 1 LSB
		data_response[4] = 0xAA;			//Ath Name byte 1 first char
				
		manuf_data_response.company_identifier = BLE_MANUF_ID_ULB;
		manuf_data_response.data.p_data        = data_response;
		manuf_data_response.data.size          = 22;
				
		//populate the scan response packet
		memset(&advdata_response, 0, sizeof(advdata_response)); //initialize all structs to 0
		advdata_response.name_type             = BLE_ADVDATA_NO_NAME;
		advdata_response.p_manuf_specific_data = &manuf_data_response;
		
		
		//set the advertisement and scan response packets
    err_code = ble_advdata_set(&advdata, &advdata_response);
    APP_ERROR_CHECK(err_code);
}
*/


//TAG SERVICE LED CHARACTERISTIC WRITE HANDLER
//This is the function that will be called when a BLE client writes to control the LED state
static void led_write_handler(ble_tag_t *p_tag, uint8_t led_state)
{
	 if(led_state)
	 {
		  nrf_gpio_pin_clear(SF_TAG_LED);  //turn ON the led
	 }
	 else
	 {
		  nrf_gpio_pin_set(SF_TAG_LED);  //turn OFF the led
	 }
}


//TAG SERVICE INITIALIZATION
static void services_init(void)
{
    uint32_t 				err_code;
	  ble_tag_init_t 	init;
	
	  init.led_write_handler = led_write_handler;
	
	  err_code = ble_tag_init(&m_tag, &init);
		APP_ERROR_CHECK(err_code);
	  RTT_PRINTF("Init TAG Service: UUID_type = %d\n", m_tag.uuid_type);
	
	  
}


/*  Function for initializing security parameters.
 */
static void sec_params_init(void)
{
    m_sec_params.bond         = SEC_PARAM_BOND;
    m_sec_params.mitm         = SEC_PARAM_MITM;
    m_sec_params.io_caps      = SEC_PARAM_IO_CAPABILITIES;
    m_sec_params.oob          = SEC_PARAM_OOB;
    m_sec_params.min_key_size = SEC_PARAM_MIN_KEY_SIZE;
    m_sec_params.max_key_size = SEC_PARAM_MAX_KEY_SIZE;
}


/* Function for handling the Connection Parameters Module.
 *
 * This function will be called for all events in the Connection Parameters Module which
 *          are passed to the application.
 *          @note All this function does is to disconnect. This could have been done by simply
 *                setting the disconnect_on_fail config parameter, but instead we use the event
 *                handler mechanism to demonstrate its use.
 *
 * @param[in]   p_evt   Event received from the Connection Parameters Module.
 */
static void on_conn_params_evt(ble_conn_params_evt_t * p_evt)
{
    uint32_t err_code;

    if(p_evt->evt_type == BLE_CONN_PARAMS_EVT_FAILED)
    {
        err_code = sd_ble_gap_disconnect(m_conn_handle, BLE_HCI_CONN_INTERVAL_UNACCEPTABLE);
        APP_ERROR_CHECK(err_code);
    }
}


/**@brief Function for handling a Connection Parameters error.
 *
 * @param[in]   nrf_error   Error code containing information about what went wrong.
 */
static void conn_params_error_handler(uint32_t nrf_error)
{
    APP_ERROR_HANDLER(nrf_error);
}


/**@brief Function for initializing the Connection Parameters module.
 */
static void conn_params_init(void)
{
    uint32_t               err_code;
    ble_conn_params_init_t cp_init;

    memset(&cp_init, 0, sizeof(cp_init));

    cp_init.p_conn_params                  = NULL;
    cp_init.first_conn_params_update_delay = FIRST_CONN_PARAMS_UPDATE_DELAY;
    cp_init.next_conn_params_update_delay  = NEXT_CONN_PARAMS_UPDATE_DELAY;
    cp_init.max_conn_params_update_count   = MAX_CONN_PARAMS_UPDATE_COUNT;
    cp_init.start_on_notify_cccd_handle    = BLE_GATT_HANDLE_INVALID;
    cp_init.disconnect_on_fail             = false;
    cp_init.evt_handler                    = on_conn_params_evt;
    cp_init.error_handler                  = conn_params_error_handler;

    err_code = ble_conn_params_init(&cp_init);
    APP_ERROR_CHECK(err_code);
}


/**@brief Function for starting timers.
*/
static void timers_start(void)
{
    uint32_t err_code;
		
		//This is where we start the LED blinking timer
    err_code = app_timer_start(m_led_blink_timer_id, LED_BLINK_OFF_TIME, NULL);
    APP_ERROR_CHECK(err_code); 	 
}


/*  Function for starting advertising.
 */
static void advertising_start(void)
{
    uint32_t             err_code;
    ble_gap_adv_params_t adv_params;

    // Start advertising
    memset(&adv_params, 0, sizeof(adv_params));

    adv_params.type        = BLE_GAP_ADV_TYPE_ADV_IND;
    adv_params.p_peer_addr = NULL;
    adv_params.fp          = BLE_GAP_ADV_FP_ANY;
    adv_params.interval    = APP_ADV_INTERVAL;
    adv_params.timeout     = APP_ADV_TIMEOUT_IN_SECONDS;

    err_code = sd_ble_gap_adv_start(&adv_params);
    APP_ERROR_CHECK(err_code);

}


/* Function for handling the Application's BLE Stack events.
 *
 * param[in]   p_ble_evt   Bluetooth stack event.
 */
static void on_ble_evt(ble_evt_t *p_ble_evt)
{
    uint32_t                         err_code;
    static ble_gap_evt_auth_status_t m_auth_status;
    bool                             master_id_matches;
    ble_gap_sec_kdist_t *            p_distributed_keys;
    ble_gap_enc_info_t *             p_enc_info;
    ble_gap_irk_t *                  p_id_info;
    ble_gap_sign_info_t *            p_sign_info;

    static ble_gap_enc_key_t         m_enc_key;           /**< Encryption Key (Encryption Info and Master ID). */
    static ble_gap_id_key_t          m_id_key;            /**< Identity Key (IRK and address). */
    static ble_gap_sign_info_t       m_sign_key;          /**< Signing Key (Connection Signature Resolving Key). */
    static ble_gap_sec_keyset_t      m_keys = {.keys_periph = {&m_enc_key, &m_id_key, &m_sign_key}};

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            m_conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
				    SEGGER_RTT_WriteString(0, "Connect\n");
            break;

        case BLE_GAP_EVT_DISCONNECTED:
            m_conn_handle = BLE_CONN_HANDLE_INVALID;
						SEGGER_RTT_WriteString(0, "Disconnect\n");
            advertising_start();
				    SEGGER_RTT_WriteString(0, "Advertising...\n");
            break;

        case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
            err_code = sd_ble_gap_sec_params_reply(m_conn_handle,
                                                   BLE_GAP_SEC_STATUS_SUCCESS,
                                                   &m_sec_params,
                                                   &m_keys);
            APP_ERROR_CHECK(err_code);
            break;

        case BLE_GATTS_EVT_SYS_ATTR_MISSING:
            err_code = sd_ble_gatts_sys_attr_set(m_conn_handle,
                                                 NULL,
                                                 0,
                                                 BLE_GATTS_SYS_ATTR_FLAG_SYS_SRVCS | BLE_GATTS_SYS_ATTR_FLAG_USR_SRVCS);
            APP_ERROR_CHECK(err_code);
            break;

        case BLE_GAP_EVT_AUTH_STATUS:
            m_auth_status = p_ble_evt->evt.gap_evt.params.auth_status;
            break;

        case BLE_GAP_EVT_SEC_INFO_REQUEST:
            master_id_matches  = memcmp(&p_ble_evt->evt.gap_evt.params.sec_info_request.master_id,
                                        &m_enc_key.master_id,
                                        sizeof(ble_gap_master_id_t)) == 0;
            p_distributed_keys = &m_auth_status.kdist_periph;

            p_enc_info  = (p_distributed_keys->enc  && master_id_matches) ? &m_enc_key.enc_info : NULL;
            p_id_info   = (p_distributed_keys->id   && master_id_matches) ? &m_id_key.id_info   : NULL;
            p_sign_info = (p_distributed_keys->sign && master_id_matches) ? &m_sign_key         : NULL;

            err_code = sd_ble_gap_sec_info_reply(m_conn_handle, p_enc_info, p_id_info, p_sign_info);
            APP_ERROR_CHECK(err_code);
            break;

        case BLE_GAP_EVT_TIMEOUT:
            if (p_ble_evt->evt.gap_evt.params.timeout.src == BLE_GAP_TIMEOUT_SRC_ADVERTISING)
            {
                // Go to system-off mode (this function will not return; wakeup will cause a reset)
								SEGGER_RTT_WriteString(0, "Adv Timeout\n");
							  err_code = app_gpiote_user_disable(m_gpiote_user_1);
							  nrf_gpio_pin_set(LED_4); //Set SFST RB0 LOW before shutdown
							  SEGGER_RTT_WriteString(0, "Going to sleep\n");
							  nrf_delay_ms(1000);
                err_code = sd_power_system_off();   
                APP_ERROR_CHECK(err_code);
            }
            break;

        default:
            // No implementation needed.
            break;
    }
}


/* Function for dispatching a BLE stack event to all modules with a BLE stack event handler.
 *
 * This function is called from the scheduler in the main loop after a BLE stack
 * event has been received.
 *
 * param[in]   p_ble_evt   Bluetooth stack event.
 */
static void ble_evt_dispatch(ble_evt_t *p_ble_evt)
{
    on_ble_evt(p_ble_evt);
    ble_conn_params_on_ble_evt(p_ble_evt);
	
	  //TAG SERVICE EVENT HANDLER
	  ble_tag_on_ble_evt(&m_tag, p_ble_evt);  //Register the Event handler function for our TAG Service to recieve BLE events
   
}


/**@brief Function for dispatching a system events to interested modules.
 *
 * @details This function is called from the System event interrupt handler after a system
 *          event has been received.
 *
 * @param[in]   sys_evt   System stack event.
 */
static void sys_evt_dispatch(uint32_t sys_evt)
{
    
}


/**@brief Function for initializing the BLE stack.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 */
static void ble_stack_init(void)
{
    uint32_t err_code;

    // Initialize the SoftDevice handler module.
    SOFTDEVICE_HANDLER_INIT(NRF_CLOCK_LFCLKSRC_XTAL_20_PPM, NULL);

#ifdef S110
    // Enable BLE stack 
    ble_enable_params_t ble_enable_params;
    memset(&ble_enable_params, 0, sizeof(ble_enable_params));
    ble_enable_params.gatts_enable_params.service_changed = IS_SRVC_CHANGED_CHARACT_PRESENT;
    err_code = sd_ble_enable(&ble_enable_params);
    APP_ERROR_CHECK(err_code);
#endif
    
    // Register with the SoftDevice handler module for BLE events.
    err_code = softdevice_ble_evt_handler_set(ble_evt_dispatch);
    APP_ERROR_CHECK(err_code);
    
    // Register with the SoftDevice handler module for System events.
    err_code = softdevice_sys_evt_handler_set(sys_evt_dispatch);
    APP_ERROR_CHECK(err_code);
}


/**@brief Function for the Event Scheduler initialization.
 */
static void scheduler_init(void)
{
    APP_SCHED_INIT(SCHED_MAX_EVENT_DATA_SIZE, SCHED_QUEUE_SIZE);
}


/** Function for initializing the GPIOTE handler module.
 */
static void gpiote_init(void)
{
   uint32_t low_to_high_bitmask = BTN_1_BITMASK | BTN_2_BITMASK; //watch for changes on Button 1 and Button 2
   uint32_t high_to_low_bitmask = BTN_1_BITMASK | BTN_2_BITMASK;
	 uint32_t retval;
	
	 APP_GPIOTE_INIT(APP_GPIOTE_MAX_USERS);
	
	 retval = app_gpiote_user_register(&m_gpiote_user_1,
																			low_to_high_bitmask,
																			high_to_low_bitmask,
																			handle_gpiote_event);
	 if(retval != NRF_SUCCESS)
	 {
		 
	 }
	 
	 retval = app_gpiote_user_enable(m_gpiote_user_1);
	 if(retval != NRF_SUCCESS)
	 {
		 
	 }
	
	
}




/*  Function for the Power manager.
 */
static void power_manage(void)
{
    uint32_t err_code = sd_app_evt_wait();
    APP_ERROR_CHECK(err_code);
}




/*  Function for application main entry.
 */
int main(void)
{
    
   
	SEGGER_RTT_WriteString(0, "TAG Reset\n");
	// Initialize
	
	  gpio_init();
	  nrf_gpio_pin_set(10);
	
	
    timers_init();
	  ble_stack_init();
    scheduler_init();
    gap_params_init();
    advertising_init();
    services_init();
    conn_params_init();
    sec_params_init();
	  gpiote_init();

    // Start execution
    timers_start();
    advertising_start();
		SEGGER_RTT_WriteString(0, "Advertising...\n");
    
		// Enter main loop
    for (;;)
    {
        app_sched_execute();
        power_manage();
    }
}


